/*++
 
// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF
// ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
// PARTICULAR PURPOSE.

Module Name:

    Device.cpp

Abstract:

    This module contains the implementation for the Badge Board driver
    device callback object.

--*/

#include "internal.h"
#include "SensorDriver.h"
#include "Device.h"
#include "Queue.h"

#include "Device.tmh"

#include <setupapi.h>
#include <devpkey.h>
#include <strsafe.h>

WCHAR g_wszDeviceName[MAX_PATH] = {0};

/////////////////////////////////////////////////////////////////////////
//
// CMyDevice::CMyDevice
//
// Object constructor function
//
/////////////////////////////////////////////////////////////////////////
CMyDevice::CMyDevice() :
    m_pWdfDevice(NULL),
    m_pBadgeSensorDriver(NULL)
{

}

/////////////////////////////////////////////////////////////////////////
//
// CMyDevice::~CMyDevice
//
// Object destructor function
//
/////////////////////////////////////////////////////////////////////////
CMyDevice::~CMyDevice()
{
    SAFE_RELEASE(m_pBadgeSensorDriver);
}

/////////////////////////////////////////////////////////////////////////
//
// CMyDevice::CreateInstance
//
// This static method is used to create and initialize an instance of
// CMyDevice for use with a given hardware device.
//
// Parameters:
//      
//      pDeviceInit  - pointer to an interface used to intialize the device
//      pDriver      - pointer to an IWDFDriver interface
//
// Return Values:
//      S_OK: object created successfully
//
/////////////////////////////////////////////////////////////////////////
HRESULT CMyDevice::CreateInstance(
    __in  IWDFDriver*           pDriver,
    __in  IWDFDeviceInitialize* pDeviceInit
    )
{
    CComObject<CMyDevice>* pMyDevice = NULL;

    HRESULT hr = CComObject<CMyDevice>::CreateInstance(&pMyDevice);

    if (SUCCEEDED(hr))
    {
	    Trace(TRACE_LEVEL_INFORMATION, "%!FUNC! Entry");

        pMyDevice->AddRef();
    
        // Prepare device parameters
        pDeviceInit->SetLockingConstraint(WdfDeviceLevel);  // The skeleton uses device level locking
        pDeviceInit->SetPowerPolicyOwnership(FALSE);        // Power policy
        // pDeviceInit->SetFilter();                        // If you're writing a filter driver then set this flag
        // pDeviceInit->AutoForwardCreateCleanupClose(WdfTrue);

        CComPtr<IUnknown> spCallback;
        hr = pMyDevice->QueryInterface(IID_IUnknown, (void**)&spCallback);

        CComPtr<IWDFDevice> spIWDFDevice;
        if (SUCCEEDED(hr))
        {
            // Create the IWDFDevice object
            hr = pDriver->CreateDevice(pDeviceInit, spCallback, &spIWDFDevice);
        }
    }

    return hr;
}

/////////////////////////////////////////////////////////////////////////
//
// CMyDevice::OnPrepareHardware
//
//  Called by UMDF to prepare the hardware for use. In our case
//  we create the SensorDDI object and initialize the Sensor Class Extension
//
// Parameters:
//      pWdfDevice - pointer to an IWDFDevice object representing the
//      device
//
// Return Values:
//      S_OK: success
//
/////////////////////////////////////////////////////////////////////////
HRESULT CMyDevice::OnPrepareHardware(
        __in IWDFDevice* pWdfDevice
        )
{
    Trace(TRACE_LEVEL_INFORMATION, "%!FUNC! Entry");

    HRESULT hr = (NULL != pWdfDevice) ? S_OK : E_UNEXPECTED;

    // Store the IWDFDevice pointer
    m_pWdfDevice = pWdfDevice;

    // Create & Configure the default IO Queue
    if(SUCCEEDED(hr))
    {
        hr = ConfigureQueue();
    }

    // Create the Badge Sensor Driver object
    if(SUCCEEDED(hr))
    {
        hr = CComObject<CBadgeSensorDriver>::CreateInstance(&m_pBadgeSensorDriver);
        if ((SUCCEEDED(hr)) && (NULL != m_pBadgeSensorDriver))
        {
            m_pBadgeSensorDriver->AddRef();
        }

        // Initialize the Badge Sensor Driver object
        if(SUCCEEDED(hr))
        {
            hr = m_pBadgeSensorDriver->Initialize(m_pWdfDevice);
        }
    }
    
    StringCchCopy(g_wszDeviceName, MAX_PATH, DEVICE_MODEL_VALUE);
    
    ULONG ulCchInstanceId = 0;
    WCHAR wszInstanceId[MAX_PATH] = {0};
    if (SUCCEEDED(pWdfDevice->RetrieveDeviceInstanceId(NULL, &ulCchInstanceId)))
    {
        if (SUCCEEDED(pWdfDevice->RetrieveDeviceInstanceId(wszInstanceId, &ulCchInstanceId)))
        {
            HDEVINFO hDeviceInfo = INVALID_HANDLE_VALUE;
            if (INVALID_HANDLE_VALUE != (hDeviceInfo = ::SetupDiCreateDeviceInfoList(NULL, NULL)))
            {
                SP_DEVINFO_DATA deviceInfo = {sizeof(SP_DEVINFO_DATA)};
                if (TRUE == ::SetupDiOpenDeviceInfo(hDeviceInfo, wszInstanceId, NULL, 0, &deviceInfo))
                {
                    DEVPROPTYPE propType;
                    ULONG ulSize;
                
                    g_wszDeviceName[0] = '\0';
                    ::SetupDiGetDeviceProperty(hDeviceInfo, &deviceInfo, &DEVPKEY_Device_FriendlyName, &propType, (PBYTE)g_wszDeviceName, sizeof(g_wszDeviceName), &ulSize, 0);
                }
                
                ::SetupDiDestroyDeviceInfoList(hDeviceInfo);
            }
        }
    }

    return hr;
}

/////////////////////////////////////////////////////////////////////////
//
// CMyDevice::OnReleaseHardware
//
// Called by UMDF to uninitialize the hardware.
//
// Parameters:
//      pWdfDevice - pointer to an IWDFDevice object for the device
//
// Return Values:
//      S_OK:
//
/////////////////////////////////////////////////////////////////////////
HRESULT CMyDevice::OnReleaseHardware(
        __in IWDFDevice* pWdfDevice
        )
{
    UNREFERENCED_PARAMETER(pWdfDevice);
    
    HRESULT hr = S_OK;

    // Uninitialize the Badge Sensor Driver object
    if(NULL != m_pBadgeSensorDriver)
    {
        m_pBadgeSensorDriver->Uninitialize();
        SAFE_RELEASE(m_pBadgeSensorDriver);
    }

    // Release the IWDFDevice handle, if it matches
    if (pWdfDevice == m_pWdfDevice.p)
    {
        m_pWdfDevice.Release();
    }

    return hr;
}

/////////////////////////////////////////////////////////////////////////
//
// CMyDevice::OnD0Entry
//
// This method is called after a new device enters the system
//
// Parameters:
//      pWdfDevice - pointer to a device object
// 
//
/////////////////////////////////////////////////////////////////////////
HRESULT CMyDevice::OnD0Entry(
        __in IWDFDevice* pWdfDevice,
        __in WDF_POWER_DEVICE_STATE previousState
        )
{
    Trace(TRACE_LEVEL_INFORMATION, "%!FUNC! Entry");

    UNREFERENCED_PARAMETER(pWdfDevice);
    UNREFERENCED_PARAMETER(previousState);

    HRESULT hr = S_OK;

    if( SUCCEEDED(hr) && NULL != m_pBadgeSensorDriver)
    {
        hr = m_pBadgeSensorDriver->Start();
    }

    return hr;
}

/////////////////////////////////////////////////////////////////////////
//
// CMyDevice::OnD0Exit
//
// This method is called when a device leaves the system
//
// Parameters:
//      pWdfDevice - pointer to a device object
//
// 
//
/////////////////////////////////////////////////////////////////////////
HRESULT CMyDevice::OnD0Exit(
        __in IWDFDevice* pWdfDevice,
        __in WDF_POWER_DEVICE_STATE newState
        )
{
    Trace(TRACE_LEVEL_INFORMATION, "%!FUNC! Entry");

    UNREFERENCED_PARAMETER(pWdfDevice);
    UNREFERENCED_PARAMETER(newState);

    HRESULT hr = S_OK;

    if( SUCCEEDED(hr) && NULL != m_pBadgeSensorDriver)
    {
        hr = m_pBadgeSensorDriver->Stop();
    }

    return hr;
}

VOID CMyDevice::OnSurpriseRemoval(
    __in IWDFDevice* pWdfDevice
    )
{
    UNREFERENCED_PARAMETER(pWdfDevice);
    Trace(TRACE_LEVEL_INFORMATION, "%!FUNC! Entry");

    return;
}

HRESULT CMyDevice::OnQueryRemove(
        __in IWDFDevice* pWdfDevice
        )
{
    UNREFERENCED_PARAMETER(pWdfDevice);
    Trace(TRACE_LEVEL_INFORMATION, "%!FUNC! Entry");

    return S_OK;
}

HRESULT CMyDevice::OnQueryStop(
        __in IWDFDevice* pWdfDevice
        )
{
    UNREFERENCED_PARAMETER(pWdfDevice);
    Trace(TRACE_LEVEL_INFORMATION, "%!FUNC! Entry");

    return S_OK;
}

/////////////////////////////////////////////////////////////////////////
//
// CMyDevice::OnCleanupFile
//
// This method is called when the file handle to the device is closed
//
// Parameters:
//      pWdfFile - pointer to a file object
//
/////////////////////////////////////////////////////////////////////////
VOID CMyDevice::OnCleanupFile(
            __in IWDFFile* pWdfFile
            )
{
    if (NULL != m_pBadgeSensorDriver)
    {
        m_pBadgeSensorDriver->CleanupFile(pWdfFile);
    }

    return;
}

/////////////////////////////////////////////////////////////////////////
//
// CMyDevice::ConfigureQueue
//  
//  This method is called after the device callback object has been initialized 
//  and returned to the driver.  It would setup the device's queues and their 
//  corresponding callback objects.
//
// Parameters:
//
// Return Values:
//      S_OK: success
//
/////////////////////////////////////////////////////////////////////////
HRESULT CMyDevice::ConfigureQueue()
{
    HRESULT hr = CMyQueue::CreateInstance(m_pWdfDevice, this);

    return hr;
}

/////////////////////////////////////////////////////////////////////////
//
// CMyDevice::ProcessIoControl
//
// This method is a helper that takes the incoming IOCTL and forwards
// it to the Windows Sensor Class Extension for processing.
//
// Parameters:
//      pQueue                  - [in] pointer to the UMDF queue that handled the request
//      pRequest                - [in] pointer to the request
//      ControlCode             - [in] the IOCTL code
//      InputBufferSizeInBytes  - [in] size of the incoming IOCTL buffer
//      OutputBufferSizeInBytes - [out] size of the outgoing IOCTL buffer
//      pcbWritten              - pointer to a DWORD containing the number of bytes returned
//
// Return Values:
//      S_OK:
//
/////////////////////////////////////////////////////////////////////////
HRESULT CMyDevice::ProcessIoControl(__in IWDFIoQueue*     pQueue,
                                    __in IWDFIoRequest*   pRequest,
                                    __in ULONG            ControlCode,
                                         SIZE_T           InputBufferSizeInBytes,
                                         SIZE_T           OutputBufferSizeInBytes,
                                         DWORD*           pcbWritten)
{
    UNREFERENCED_PARAMETER(pQueue);
    UNREFERENCED_PARAMETER(ControlCode);
    UNREFERENCED_PARAMETER(InputBufferSizeInBytes);
    UNREFERENCED_PARAMETER(OutputBufferSizeInBytes);
    UNREFERENCED_PARAMETER(pcbWritten);

    HRESULT hr = S_OK;
    
    if(NULL != m_pBadgeSensorDriver)
    {
        hr = m_pBadgeSensorDriver->ProcessIoControl(pRequest);
    }

    return hr;
}