/*

// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF
// ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
// PARTICULAR PURPOSE.

Module Name:
    BadgeSensorDriver.cpp

Abstract:
    Implements the CBadgeSensorDriver container class

--*/

#include "internal.h"
#include "SensorDDI.h"
#include "SensorDriver.h"
#include "SensorDevice.h"

#include "SensorDriver.tmh"

SENSOR_MAP  g_SupportedSensors;      // map of supported sensor types and their objectIds
SENSOR_LIST g_pSensorList;          // the object used by the driver to get sensor data

/////////////////////////////////////////////////////////////////////////
//
// CBadgeSensorDriver::CBadgeSensorDriver
//
// Object constructor function
//
/////////////////////////////////////////////////////////////////////////
CBadgeSensorDriver::CBadgeSensorDriver() :
    m_pWdfDevice(NULL),
    m_pClassExtension(NULL),
    m_pSensorDDI(NULL),
    m_hSensorEvent(NULL),
    m_hBadgeSensorEventingThread(NULL),
    m_fBadgeSensorDriverInitialized(FALSE),
    m_fThreadActive(FALSE)
{

}


/////////////////////////////////////////////////////////////////////////
//
// CBadgeSensorDriver::~CBadgeSensorDriver
//
// Object destructor function
//
/////////////////////////////////////////////////////////////////////////
CBadgeSensorDriver::~CBadgeSensorDriver()
{
    SAFE_RELEASE(m_pSensorDDI);
}

/////////////////////////////////////////////////////////////////////////
//
// CBadgeSensorDriver::Initialize
//
//
//
//
/////////////////////////////////////////////////////////////////////////
HRESULT CBadgeSensorDriver::Initialize(__in IWDFDevice* pWdfDevice)
{
    HRESULT hr = (FALSE == IsInitialized()) ? S_OK : E_UNEXPECTED;

    if(SUCCEEDED(hr))
    {
        g_SupportedSensors[Accelerometer]       = SENSOR_ACCELEROMETER_ID;
        g_SupportedSensors[AmbientLight]        = SENSOR_AMBIENT_LIGHT_ID;
        g_SupportedSensors[SwitchArrayLeft]     = SENSOR_SWITCH_ARRAY_LEFT_ID;
        g_SupportedSensors[SwitchArrayRight]    = SENSOR_SWITCH_ARRAY_RIGHT_ID;

        // Construct 4 BadgeSensorDevices to hold the sensors
        // This wil also Initialize the CBadgeSensorDevice with known 
        // Supported PROPERTY & DATAFIELD propertykeys 
        // Also set default values for all of them
    
        POSITION pos = g_SupportedSensors.GetStartPosition();
        while( NULL != pos)
        {
            CBadgeSensorDevice *pSensor     = new CBadgeSensorDevice();
            SensorType sType                = g_SupportedSensors.GetNextKey(pos);
            
            // Initialize every sensor
            hr = pSensor->Initialize(sType);
            
            // Set the unique persistent ID
            if(SUCCEEDED(hr))
            {
                pSensor->SetUniqueID(pWdfDevice);
            }
            
            g_pSensorList.AddTail(pSensor);
        }

        // Store the IWDF Device pointer
        m_pWdfDevice = pWdfDevice;

        // Create and initialize the Sensor Class Extension
        if(SUCCEEDED(hr))
        {
            hr = InitializeClassExtension();
        }

        if(SUCCEEDED(hr))
        {
            m_fBadgeSensorDriverInitialized = TRUE;
        }
    }

    return hr;
}

/////////////////////////////////////////////////////////////////////////
//
// CBadgeSensorDriver::Uninitialize
//
//
//
//
/////////////////////////////////////////////////////////////////////////
void CBadgeSensorDriver::Uninitialize()
{
    // Free all the sensor objects
    for(DWORD i = 0; i < g_pSensorList.GetCount(); i++)
    {
        CBadgeSensorDevice *pSensor = g_pSensorList.GetAt(g_pSensorList.FindIndex(i));

        pSensor->Uninitialize();

        delete pSensor;
    }

    // Clear the Sensor list and Sensor map
    g_pSensorList.RemoveAll();
    g_SupportedSensors.RemoveAll();

    // Release Sensor Class Extension and Sensor DDI
    if(NULL != m_pClassExtension)
    {
        m_pClassExtension->Uninitialize();
        m_pClassExtension.Release();
    }

    SAFE_RELEASE(m_pSensorDDI);

    m_fBadgeSensorDriverInitialized = FALSE;

    return;
}

/////////////////////////////////////////////////////////////////////////
//
// CBadgeSensorDriver::InitializeClassExtension
//
//
//
//
/////////////////////////////////////////////////////////////////////////
HRESULT CBadgeSensorDriver::InitializeClassExtension()
{
    HRESULT hr = (NULL == m_pClassExtension) ? S_OK : E_UNEXPECTED;
    
    if(SUCCEEDED(hr))
    {
        //
        // Create the SensorDDI object that implements ISensorDriver
        //
        hr = CComObject<CSensorDDI>::CreateInstance(&m_pSensorDDI);
        if ((SUCCEEDED(hr)) && (NULL != m_pSensorDDI))
        {
            m_pSensorDDI->AddRef();
        }

        if(SUCCEEDED(hr))
        {
            // CoCreate the Sensor ClassExtension
            if(SUCCEEDED(hr))
            {
                hr =  CoCreateInstance(CLSID_SensorClassExtension,
                                        NULL,
                                        CLSCTX_INPROC_SERVER,
                                        IID_ISensorClassExtension,
                                        (VOID**)&m_pClassExtension);
            }
    
            // Initialize Sensor ClassExtension
            if(SUCCEEDED(hr))
            {
                CComPtr<IUnknown> spIUnknown;
                hr = m_pSensorDDI->QueryInterface(IID_IUnknown, (void**)&spIUnknown);
                
                if(SUCCEEDED(hr))
                {
                    hr = m_pClassExtension->Initialize(m_pWdfDevice, spIUnknown);
                }
            }
        }
    }

    return hr;
}

/////////////////////////////////////////////////////////////////////////////////
//
// CBadgeSensorDriver::Start
//
//
//
//
//////////////////////////////////////////////////////////////////////////////////
HRESULT CBadgeSensorDriver::Start()
{
    Trace(TRACE_LEVEL_INFORMATION, "%!FUNC! Entry");

    // Make sure we are initialized
    HRESULT hr = (TRUE == IsInitialized()) ? S_OK : E_UNEXPECTED;

    // Initialize the sensor device
    if(SUCCEEDED(hr))
    {
        if(NULL != m_pSensorDDI)
        {
            hr = m_pSensorDDI->InitSensorDevice(m_pWdfDevice);
        }
        else
        {
            hr = E_UNEXPECTED;
            Trace(TRACE_LEVEL_ERROR, "SensorDDI pointer is NULL, hr = %!HRESULT!", hr);
        }
    }

    if (SUCCEEDED(hr))
    {  
        // Step 1: Create the Data Changed Event Handle
        m_hSensorEvent = ::CreateEvent(  NULL,        // No security attributes
                                        FALSE,       // Automatic-reset event object
                                        FALSE,       // Initial state is non-signaled
                                        NULL  );     // Unnamed object
        
     
        
        for(DWORD i = 0; i < g_pSensorList.GetCount(); i++)
        {
            g_pSensorList.GetAt(g_pSensorList.FindIndex(i))->SetDataEventHandle(m_hSensorEvent);
        }

        // Step 2: Activate & Create and start the eventing thread
        Activate();

        m_hBadgeSensorEventingThread = ::CreateThread(NULL,                                              // Cannot be inherited by child process
                                                     0,                                                   // Default stack size
                                                     &CBadgeSensorDriver::_SensorEventThreadProc,   // Thread proc
                                                     (LPVOID)this,                                        // Thread proc argument
                                                     0,                                                   // Starting state = running
                                                     NULL);                                               // No thread identifier

    }

    return hr;
}

/////////////////////////////////////////////////////////////////////////////////
//
// CBadgeSensorDriver::Stop
//
//
//
//
//////////////////////////////////////////////////////////////////////////////////
HRESULT CBadgeSensorDriver::Stop()
{
    Trace(TRACE_LEVEL_INFORMATION, "%!FUNC! Entry");

    HRESULT hr = S_OK;

    // DeInitialize the sensor device
    if(NULL != m_pSensorDDI)
    {
        hr = m_pSensorDDI->DeInitSensorDevice();
    }

    // Step 1: Stop the Eventing thread and Close the handle
    if (NULL != m_hBadgeSensorEventingThread)
    {
        // De-activate and close the thread
        DeActivate();
        WaitForSingleObject(m_hBadgeSensorEventingThread, INFINITE);
        CloseHandle(m_hBadgeSensorEventingThread);
    }
   
    // Step 2: Close the Data Change Event handle
    if (NULL != m_hSensorEvent)
    {
        CloseHandle(m_hSensorEvent);
    }

    return hr;

}

/////////////////////////////////////////////////////////////////////////////////
//
// CBadgeSensorDriver::_SensorEventThreadProc
//
//
//
//
//////////////////////////////////////////////////////////////////////////////////
DWORD WINAPI CBadgeSensorDriver::_SensorEventThreadProc(__in LPVOID pvData)
{
    CBadgeSensorDriver* pParent = (CBadgeSensorDriver*)pvData;

    if (NULL == pParent)
    {
        return 0;
    }

    HRESULT hr = CoInitializeEx(NULL, COINIT_MULTITHREADED);

    if (FAILED(hr))
    {
        Trace(TRACE_LEVEL_ERROR, "Failed to call CoInitialize on _SensorEventThreadProc thread, hr = %!HRESULT!", hr);
        return 0;
    }
    
    // Create the event parameters collection if it doesn't exist
    CComPtr<IPortableDeviceValues> pEventParams;
    if (pEventParams == NULL)
    {
        hr = CoCreateInstance(CLSID_PortableDeviceValues,
                              NULL,
                              CLSCTX_INPROC_SERVER,
                              IID_IPortableDeviceValues,
                              (VOID**)&pEventParams);
    }

    if (FAILED(hr))
    {
        Trace(TRACE_LEVEL_ERROR, "Failed to CoCreateInstance for Event Parameters, hr = %!HRESULT!", hr);
        return 0;
    }
    
    while (pParent->IsActive() &&
           (WAIT_OBJECT_0 == WaitForSingleObject(pParent->GetSensorEventHandle(), INFINITE)))
    {    
        HRESULT hr = S_OK;

        if (S_OK == hr)
        {
            // Initialize the event parameters
            pEventParams->Clear();

            // Populate the event type
            hr = pEventParams->SetGuidValue(SENSOR_EVENT_PARAMETER_EVENT_ID, SENSOR_EVENT_DATA_UPDATED);
        }


        // Loop through every sensor and post an event for each one if needed
        if (S_OK == hr)
        {
            for(DWORD i = 0; i < g_pSensorList.GetCount(); i++)
            {
                CBadgeSensorDevice *pSensor;
                CComBSTR objectId;
                
                pSensor = g_pSensorList.GetAt(g_pSensorList.FindIndex(i));
                
                // Check if this Sensor has valid data to post
                if( TRUE == pSensor->HasValidDataEvent())
                {
                    // Get the All the Data Field values
                    // Populate the event parameters
                    hr = pSensor->GetAllDataFieldValues(pEventParams);

                    if( SUCCEEDED(hr))
                    {
                        objectId = pSensor->GetSensorObjectID();
                        pParent->PostDataEvent(objectId, pEventParams);
                    }
                }
                 
            }
        }
   
    }
    
    CoUninitialize();

    return 0;
}

/////////////////////////////////////////////////////////////////////////////////
//
// CBadgeSensorDriver::PostDataEvent
//
//
//
//
//////////////////////////////////////////////////////////////////////////////////
HRESULT CBadgeSensorDriver::PostDataEvent(__in LPWSTR objectId, IPortableDeviceValues *pValues)
{
    Trace(TRACE_LEVEL_INFORMATION, "%!FUNC! Entry");

    HRESULT hr = S_OK;

    CComPtr<IPortableDeviceValuesCollection> pEventCollection;
    
    if(pEventCollection == NULL)
    {
        //*--  CoCreate a collection to store the sensor object identifiers.
        hr = CoCreateInstance( CLSID_PortableDeviceValuesCollection,
                              NULL,
                              CLSCTX_INPROC_SERVER,
                              IID_IPortableDeviceValuesCollection,
                              (VOID**) &pEventCollection);
    }

    if( SUCCEEDED(hr) )
    {
        hr = pEventCollection->Add( pValues );
        if( SUCCEEDED(hr) && (m_pClassExtension != NULL) )
        {
            hr = m_pClassExtension->PostEvent( objectId, pEventCollection );            
        }
    }

    return hr;
}

/////////////////////////////////////////////////////////////////////////
//
// CBadgeSensorDriver::OnCleanupFile
//
// This method is called when the file handle to the device is closed
//
// Parameters:
//      pWdfFile - pointer to a file object
//
/////////////////////////////////////////////////////////////////////////
void CBadgeSensorDriver::CleanupFile(
    __in IWDFFile* pWdfFile
    )
{
    if (NULL != m_pClassExtension)
    {
        m_pClassExtension->CleanupFile(pWdfFile);
    }

    return;
}

/////////////////////////////////////////////////////////////////////////
//
// CBadgeSensorDriver::ProcessIoControl
//
// This method is called to process a Device IO Control
//
// Parameters:
//      pRequest - [in] pointer to the request
//
/////////////////////////////////////////////////////////////////////////
HRESULT CBadgeSensorDriver::ProcessIoControl(
    __in IWDFIoRequest*   pRequest
    )
{
    Trace(TRACE_LEVEL_INFORMATION, "%!FUNC! Entry");

    HRESULT hr = S_OK;

    if (NULL != m_pClassExtension)
    {
        hr = m_pClassExtension->ProcessIoControl(pRequest);
    }

    return hr;
}

VOID CBadgeSensorDriver::DeActivate()
{
    CComCritSecLock<CComAutoCriticalSection> scopeLock(m_CriticalSection);

    m_fThreadActive = FALSE; 
    SetEvent(m_hSensorEvent);
    return;
}

VOID CBadgeSensorDriver::Activate()
{
    CComCritSecLock<CComAutoCriticalSection> scopeLock(m_CriticalSection);

    m_fThreadActive = TRUE; 

    return;
}


